Object subclass: #Simulation
	instanceVariableNames: 'time seqSet current stop priority '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Discrete_Simulation'!
Simulation comment:
'This class implements the simulation environment (i.e. sequencing set of processes)
as used in Simula-67 programing language. 


(C) 	Vojtech Merunka
	Smetanova 647
	CZ-286 01 CASLAV

	merunka@pef.czu.cz
	http://omega.pef.czu.cz/pef/ki/merunka'!


!Simulation methodsFor: 'querying'!

current
	"Answer the current process."

	^current!

next
	"answer next process in sequencing set"

	seqSet isEmpty ifTrue: [^nil].
	^seqSet at: 1!

nextIfNone: aBlock 
	"answer next process in sequencing set"

	seqSet isEmpty ifTrue: [^aBlock value].
	^seqSet at: 1!

nextTime
	"answer time of next process in sequencing set"

	seqSet isEmpty ifTrue: [^nil].
	^(seqSet at: 1) time!

nextTimeIfNone: aBlock
	"answer time of next process in sequencing set"

	seqSet isEmpty ifTrue: [^aBlock value].
	^(seqSet at: 1) time!

priority
	"Answer the value of priority."

	^priority!

time
	"Answer the value of time."

	^time! !

!Simulation methodsFor: 'settings'!

priority: aValue 
	"Set the value of priority.
 (default level is 3)"

	priority := aValue!

time: aValue 
	"Set the value of time."

	time := aValue! !

!Simulation methodsFor: 'initialization'!

initialize
	"private - create new simulation"

	stop := true.
	time := 0.
	priority := Processor activePriority.
	seqSet := (SortedCollection new: 0)
				sortBlock: [:e1 :e2 | e1 time < e2 time]! !

!Simulation methodsFor: 'simulation control'!

at: aNumber 
	"active process is reactivated at aNumber time"

	current time: aNumber.
	seqSet add: current.
	current stop!

delay: aNumber 
	"active process is reactivated
 by aNumber delay time"

	current time: self time + aNumber.
	seqSet add: current.
	current stop!

go
	"start or restart of simulation"

	
	[stop := false.
	[seqSet isEmpty | stop]
		whileFalse: 
			[time := (current := seqSet removeFirst) time.
			current activate].
	current := nil.
	stop := true] 
		forkAt: priority!

passivate
	"passivate the active process"

	current simulation: nil; stop!

stop
	"stop the simulation"

	stop := true!

wait: aSet 
	"passivate and store the active process
 into aSet"

	aSet add: current.
	current stop! !

!Simulation methodsFor: 'processing'!

activate: aProcess 
	"activate the process"

	seqSet add: (aProcess time: time; simulation: self)!

activate: aProcess at: aNumber 
	"activate the process at time"

	seqSet add: (aProcess time: aNumber; simulation: self)!

activate: aProcess delay: aNumber 
	"activate the process by delay time"

	seqSet add: (aProcess time: aNumber + time; simulation: self)!

cancel: aProcess 
	"active process is passivated,
 inactive process is terminated"

	aProcess = current
		ifTrue: [self passivate]
		ifFalse: 
			[seqSet remove: aProcess.
			aProcess terminate]! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Simulation class
	instanceVariableNames: ''!


!Simulation class methodsFor: 'instance creation'!

new
	"inicializace"

	^super new initialize! !

Object subclass: #SimProcess
	instanceVariableNames: 'simulation evTime terminated semaphore '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Discrete_Simulation'!
SimProcess comment:
'This class implements the common behaviour of simulation processes. The default method describing
its activity is "process". It is expected to have subclasses of this class for concrete solutions.

Sumulation environment is stored in instance variable "simulation" and may be also
accessed via mesage as "self simulation".


(C) 	Vojtech Merunka
	Smetanova 647
	CZ-286 01 CASLAV

	merunka@pef.czu.cz
	http://omega.pef.czu.cz/pef/ki/merunka'!


!SimProcess methodsFor: 'private - settings'!

simulation: aValue 
	"Set the simulation."

	simulation := aValue!

time: aValue 
	"private - Set the value of evTime."

	evTime := aValue! !

!SimProcess methodsFor: 'querying'!

simulation
	"Answer the simulation.
 If nil, process is passive"

	^simulation!

terminated
	"Answer if the receiver process
 is terminated."

	^terminated!

time
	"Answer the value of evTime."

	^evTime! !

!SimProcess methodsFor: 'initialization'!

initialize
	"private - create new process"

	terminated := false! !

!SimProcess methodsFor: 'processing'!

activate
	"private - run the process"

	semaphore isNil
		ifTrue: 
			[semaphore := Semaphore new.
			[self process]
				forkAt: simulation priority + 1.
			terminated := true]
		ifFalse: [semaphore signal]!

process
	"this must be implemented by subclass"

	self error: 'This message should be implemented in subclass!!'!

stop
	"private - stop the process"

	semaphore wait!

terminate
	"private - terminate the process"

	terminated := true.
	simulation := nil! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

SimProcess class
	instanceVariableNames: ''!


!SimProcess class methodsFor: 'instance creation'!

new
	"initializace"

	^super new initialize! !

SimProcess subclass: #Prime
	instanceVariableNames: 'number '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Discrete_Simulation'!
Prime comment:
'This class implements the example of simulation package use.

See instance metod "process".

(C) 	Vojtech Merunka
	Smetanova 647
	CZ-286 01 CASLAV

	merunka@pef.czu.cz
	http://omega.pef.czu.cz/pef/ki/merunka'!


!Prime methodsFor: 'processing'!

process
	"Eratosthenes' net. This is the main process 
	
	Transcript clear.
	(Simulation new) activate: Prime new at: 2 ; go 
	
	"

	Transcript show: (number := self time) printString , ' '.
	[self time < 1000]
		whileTrue: 
			[(self simulation next isNil or: [self simulation nextTime > (self time + 1)])
				ifTrue: [self simulation activate: Prime new delay: 1].
			simulation delay: number]! !

